import java.io.*;
import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import javax.microedition.io.*;

public class ConnectForm extends Form implements CommandListener  
{ 
    Connection     connection;
    MobiBrowser    browser;
    Thread         thread;
    boolean        stopped;
    Gauge          gauge;
    String         url;
    int            pos;
    String         label;

    static final int INIT_BUF_SIZE = 32*1024;
    static final int BUF_SIZE_EXTENT = 32*1024;

    static boolean hasFileConnection;

    static { 
        try { 
            Class.forName("javax/microedition/io/file/FileConnection");
            hasFileConnection = true;
        } catch (ClassNotFoundException x) {}
    }
    
    class InternetConnection extends Thread { 
        public void run() { 
            Connection c = null;
            InputStream is = null;
            
            try {
                try {
                    //System.out.println("Connect to " + url);
                    int len = -1;
                    if (url.startsWith("jar://")) {                         
                        is = this.getClass().getResourceAsStream(url.substring(6));
                    } else if (url.startsWith("file://") && !hasFileConnection) {
                        is = this.getClass().getResourceAsStream(url);
                    } else { 
                        c = Connector.open(url);
                        connection = c;
                        
                        if (c instanceof HttpConnection) { 
                            HttpConnection hc = (HttpConnection)c;
                            // Set the request method and headers
                            hc.setRequestMethod(HttpConnection.GET);
                            hc.setRequestProperty("User-Agent", "Profile/MIDP-1.0 Configuration/CLDC-1.0");
//                            hc.setRequestProperty("Accept","*/*");
                            hc.setRequestProperty("Accept-charset", "windows-1251, win. x-cp1251, cp1251, cp-1251"); 
                            hc.setRequestProperty("Content-Language", Locale.LANGUAGE_CODE);                    
                            // c.setRequestProperty("Content-Language", System.getProperty("microedition.locale"));
                            
                            // Opening the InputStream will open the connection
                            // and read the HTTP headers. They are stored until
                            // requested.
                            is = hc.openInputStream();
                            if (stopped) { 
                                return;
                            }
                            
                            // Get the ContentType
                            String type = hc.getType();
                            //System.out.println("Connection type=" + type);
                            //setTicker(null);
                            
                            // Get the length and process the data
                            len = (int)hc.getLength();
                            //System.out.println("Content length =" + len);
                        } else { 
                            is = ((InputConnection)c).openInputStream();
                        }
                    }
                    byte[] body;
                    if (c != null && "com.mot.cldc.io.j2me.http.Protocol".equals(c.getClass().getName())) { 
                        // Motorola implementation of connection class is something very 
                        // strage having no relation to standard J2ME connections behavoir
                        if (len >= 0) {
                            gauge.setMaxValue(len);
                            body = new byte[len+4];
                            int offs = 0;
                            while (offs < len) { 
                                int rc = is.read(body, offs, len - offs);
                                if (stopped) { 
                                    return;
                                }
                                if (rc >= 0) { 
                                    offs += rc;
                                    gauge.setValue(offs);
                                } else { 
                                    browser.error(Locale.CONNECTION_FAILURE);
                                    return;
                                }
                            }
                        } else {
                            int rc;
                            body = new byte[INIT_BUF_SIZE];

                            for (len = 0; !stopped && (rc = is.read(body, len, body.length-len-4)) >= 0;) {
                                if (rc == 1) { 
                                    if (body[len] == -1) { 
                                        break;
                                    }
                                }
                                len += rc;
                                if (len+4 == body.length) { 
                                    int newLen = (len+4)*2 < BUF_SIZE_EXTENT ? (len+4)*2 : len + 4 + BUF_SIZE_EXTENT;
                                    byte[] newBody = new byte[newLen];
                                    System.arraycopy(body, 0, newBody, 0, len);
                                    body = newBody;
                                }
                            }
                        }
                    } else { 
                        // normal implementation
                        if (len >= 0) {
                            gauge.setMaxValue(len);
                            body = new byte[len];
                            int offs = 0;
                            while (offs < len) { 
                                int rc = is.read(body, offs,len - offs);
                                if (stopped) { 
                                    return;
                                }
                                if (rc >= 0) { 
                                    offs += rc;
                                    gauge.setValue(offs);
                                } else { 
                                    browser.error(Locale.CONNECTION_FAILURE);
                                    return;
                                }
                            }
                        } else {
                            int rc;
                            body = new byte[INIT_BUF_SIZE];
                            for (len = 0; !stopped && (rc = is.read(body, len, body.length-len)) >= 0;) {
                                len += rc;
                                if (len == body.length) { 
                                    int newLen = len*2 < BUF_SIZE_EXTENT ? len*2 : len + BUF_SIZE_EXTENT;
                                    byte[] newBody = new byte[newLen];
                                    System.arraycopy(body, 0, newBody, 0, len);
                                    body = newBody;
                                }
                            }
                        }
                    }
                    if (!stopped) {
                        new ReaderCanvas(url, pos, label, body, len, browser);
                    }
                } finally { 
                    if (is != null) { 
                        is.close();
                    }
                    if (c != null) { 
                        c.close();
                    }
                } 
            } catch (Exception x) { 
                x.printStackTrace();
                System.out.println("Exception " + x);
                browser.error(Locale.CONNECTION_FAILURE);
            }
        }
    }

    ConnectForm(String url, int pos, MobiBrowser browser) { 
        super(MobiBrowser.removeHttpPrefix(url));
        url = url.indexOf("://") > 0 ? url : "http://" + url;
        this.pos = pos;
        this.browser = browser;
        int labelPos = url.lastIndexOf('#');
        if (labelPos >= 0) { 
            this.url = url.substring(0, labelPos);
            label = url.substring(labelPos+1);
        } else { 
            this.url = url;
            label = null;
        }
        setCommandListener(this);
        addCommand(MobiBrowser.STOP_CMD);
        setTicker(new Ticker(Locale.CONNECTING + MobiBrowser.removeHttpPrefix(url)));
        gauge = new Gauge(null, false, 10, 0);
        append(gauge);
        Display.getDisplay(browser).setCurrent(this);
        thread = new InternetConnection();
        thread.start();
    }
 
    public void commandAction(Command c, Displayable d)  
    {
        if (c == MobiBrowser.STOP_CMD)  { 
            stopped = true;
            Connection con = connection;
            if (con != null) {
                try { 
                    con.close();
                } catch (IOException x) {}
            }
        }
        Display.getDisplay(browser).setCurrent(browser.menu);
    }
}
        
