import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.util.*;
import java.io.*;
import javax.microedition.rms.*;

public class MobiBrowser extends MIDlet
{
    static final Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 1);
    static final Command STOP_CMD = new Command(Locale.STOP, Command.BACK, 1);
    static final Command QUIT_CMD = new Command(Locale.QUIT, Command.CANCEL, 10);
    static final Command GOTO_CMD = new Command(Locale.GOTO, Command.SCREEN, 1);
    static final Command ADD_CMD = new Command(Locale.ADD, Command.SCREEN, 2);
    static final Command RENAME_CMD = new Command(Locale.RENAME, Command.SCREEN, 1);
    static final Command REMOVE_CMD = new Command(Locale.REMOVE, Command.SCREEN, 2);

    static final Command URL_CMD = new Command(Locale.URL, Command.SCREEN, 1);
    static final Command FORWARD_CMD = new Command(Locale.FORWARD, Command.SCREEN, 2);
    static final Command BACKWARD_CMD = new Command(Locale.BACKWARD, Command.SCREEN, 3);
    static final Command BOOKMARKS_CMD = new Command(Locale.BOOKMARKS, Command.SCREEN,4);
    static final Command HISTORY_CMD = new Command(Locale.HISTORY, Command.SCREEN, 5);
    static final Command WEB_SEARCH_CMD = new Command(Locale.WEB_SEARCH, Command.SCREEN, 6);
    static final Command SEARCH_CMD = new Command(Locale.SEARCH, Command.SCREEN, 7);
    static final Command REFRESH_CMD = new Command(Locale.REFRESH, Command.SCREEN, 8);
    static final Command HELP_CMD = new Command(Locale.HELP, Command.SCREEN, 9);

    static final String STORE_NAME = "BOOKMARKS";
    static final int CACHE_SIZE_PAGES = 2;
    static final int CACHE_SIZE_BYTES  = 64*1024;

    static class CacheEntry { 
        CacheEntry   next;
        CacheEntry   prev;
        String       url;
        ReaderCanvas canvas; 

        void link(CacheEntry entry) { 
            entry.next = next;
            next.prev = entry;
            entry.prev = this;
            next = entry;
        }

        void unlink() {
            next.prev = prev;
            prev.next = next;
        }

        boolean isEmpty() { 
            return next == this;
        }
    }


    static class History { 
        String  url;
        int     pos;
        History forward;
        History backward;
    }

    public void addBookmark(Bookmark bm) { 
        for (int i = 0, n = bookmarks.size(); i < n; i++) { 
            if (bm.equals(bookmarks.elementAt(i))) { 
                Bookmark b = (Bookmark)bookmarks.elementAt(i);
                b.pos = bm.pos;
                b.url = bm.url;
                updateBookmark(b);
                return;
            }
        }
        bookmarks.addElement(bm);
        byte[] record = bm.pack();
        try { 
            bm.id = store.addRecord(record, 0, record.length);
        } catch (Exception x) {}
    }

    public void updateBookmark(Bookmark bm) { 
        byte[] record = bm.pack();
        try { 
            store.setRecord(bm.id, record, 0, record.length);
        } catch (Exception x) {}
    }        

    public void removeBookmark(int i) { 
        try { 
            store.deleteRecord(((Bookmark)bookmarks.elementAt(i)).id);
        } catch (Exception x) {}
        bookmarks.removeElementAt(i);
    }        
        

    private void openStore() {
        try { 
            store = RecordStore.openRecordStore(STORE_NAME, true);
            int last = store.getNextRecordID();
            if (last == 1) { 
                for (int i = 0; i < Locale.PREDEFINED_BOOKMARKS.length; i++) { 
                    addBookmark(Locale.PREDEFINED_BOOKMARKS[i]);
                }
            } else { 
                for (int i = 1; i < last; i++) { 
                    try { 
                        bookmarks.addElement(new Bookmark(i, store.getRecord(i)));
                    } catch (InvalidRecordIDException x) {}
                }
            }
        } catch(Exception x) {
        }
    }

    private void closeStore() {
        if (store != null) { 
            try {                
                store.closeRecordStore();
                store = null;
            } catch (Exception x) {}
        }
    }


    public static String removeHttpPrefix(String url) { 
        int prot = url.indexOf(':'); 
        return url.substring(prot+3);
    }

    public static String getHost(String url) { 
        int prot = url.indexOf(':'); 
        int host = url.indexOf('/', prot+3);
        if (host < 0) { 
            host = prot+3;
        }
        return url.substring(0, host);
    }

    public static String getBaseUrl(String url) { 
        return url.substring(0, url.lastIndexOf('/')+1);
    }

    public void load(String url, int pos) { 
        if (cacheSizePages > 0) { 
            CacheEntry entry = (CacheEntry)cacheHash.get(url);
            if (entry != null) { 
                entry.unlink();
                lru.link(entry);
                Display.getDisplay(this).setCurrent(entry.canvas);
                return;
            }
        }
        new ConnectForm(url, pos, this);
    }

    public void forward() {
        if (current != null && current.forward != null) { 
            current = current.forward;
            load(current.url, current.pos);
        }
    }

    public void backward() {
        if (current != null && current.backward != null) { 
            current = current.backward;
            load(current.url, current.pos);
        }
    }

    public void setCurrentPosition(int pos) { 
        current.pos = pos;
        addBookmark(new Bookmark(pos, Locale.LAST, current.url));
    }

    public MobiBrowser() { 
        bookmarks = new Vector();
        history = new Vector();
        historyHash = new Hashtable();
        cacheSizePages = CACHE_SIZE_PAGES;
        cacheSizeBytes = CACHE_SIZE_BYTES;
        if (cacheSizePages > 0) { 
            lru = new CacheEntry();
            lru.next = lru.prev = lru;
            cacheHash = new Hashtable();
        }
        openStore();
    }
        
    public void error(String message) { 
        error(message, menu);
    }

    public void error(String message, Displayable parent) { 
        Alert a = new Alert(Locale.ERROR, message, null, AlertType.ERROR);
        a.setTimeout(Alert.FOREVER);
        Display.getDisplay(this).setCurrent(a, parent);
    }


    protected void destroyApp(boolean unconditional) { 
        closeStore();
    }    

    protected  void pauseApp() {
    }

    protected void startApp() 
    {
        menu = new Menu(this, null);
    }

    public void quit() { 
        destroyApp(true);
        notifyDestroyed();
    }

    public void cache(String url, ReaderCanvas canvas) { 
        History h = new History();
        h.url = url;
        h.backward = current;
        if (current != null) { 
            current.forward = h;
        }
        current = h;
        history.addElement(h);
        historyHash.put(url, h);

        if (cacheSizePages > 0) { 
            CacheEntry entry = (CacheEntry)cacheHash.get(url);
            if (entry == null) { 
                cacheUsedBytes += canvas.textLength;
                cacheUsedPages += 1;
                if (!lru.isEmpty() && (cacheUsedPages > cacheSizePages || cacheUsedBytes > cacheSizeBytes)) { 
                    do { 
                        entry = lru.prev;
                        entry.unlink();
                        cacheHash.remove(entry.url);
                        cacheUsedBytes -= entry.canvas.textLength;
                        cacheUsedPages -= 1;
                        entry.canvas = null;
                    } while (!lru.isEmpty() && cacheUsedBytes > cacheSizeBytes);
                } else { 
                    entry = new CacheEntry();
                } 
                entry.url = url;
                cacheHash.put(url, entry);
            } else { 
                entry.unlink();
            }
            entry.canvas = canvas;
            lru.link(entry);
        }
    }


    Vector      bookmarks;
    Vector      history;   
    History     current;
    Hashtable   historyHash;
    RecordStore store;
    Menu        menu;
    Hashtable   cacheHash;
    CacheEntry  lru;
    int         cacheSizePages;
    int         cacheSizeBytes;
    int         cacheUsedBytes;
    int         cacheUsedPages;
}

