import javax.microedition.lcdui.*;
import java.util.*;

public class ReaderCanvas extends Canvas implements CommandListener {
    MobiBrowser browser;
    int         screenWidth;
    int         screenHeight;
    int         fontHeight;
    int         screenLines;
    int         nLines;
    int[]       linePos;
    char        prevPageChar;
    Font        font;
    Font        linkFont;
    int         spaceWidth;
    int         hyphenWidth;
    Timer       repeatTimer;
    RepeatTask  repeatTask;
    byte[]      text;
    int         textLength;
    Tag         currTag;
    Tag         tags;
    int         currPagePos;
    int         currPageEnd;
    Hashtable   index;
    String      label;
    String      url;

    static final int BACKGROUND_COLOR = 0xffffff;
    static final int FOREGROUND_COLOR = 0;
    static final int SLIDER_COLOR = 0x8080ff;
    static final int SLIDER_POS_COLOR = 0xff0000;
    static final int VISITED_LINK_COLOR = 0x800040;
    static final int LINK_COLOR = 0x0000ff;
    static final int CURRENT_LINK_BACKGROUND_COLOR = 0;
    static final int CURRENT_LINK_FOREGROUND_COLOR = 0xffffff;
    static final int MIN_SLIDER_LENGTH = 3;
    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int OVERLAP = 1;
    static final int INDENTATION = 2;

    static final String[] breakTags = {
        "li", "br", "tr", "p", "h1", "h2", "h3", "hr", "/p", "/h1", "/h2", "/h3"
    };

    static final String[] skipTags = {
        "form", "script", "style"
    };

    static class Tag { 
        Tag     next;
        Tag     prev;
        String  url;
        int     beg;
        int     end;
        int     pos;
            
        void link(Tag tag) { 
            prev.next = tag;
            tag.next = this;
            tag.prev = prev;
            prev = tag;
        }       

        void unlink() { 
            next.prev = prev;
            prev.next = next;
        }
        
        Tag(String url, int beg) {             
            this.url = url;
            this.beg = beg;
            this.end = -1;
        } 
        
        Tag() { 
            next = prev = this;
            beg = end = -1;
        }
    }

    ReaderCanvas(String url, int pos, String label, byte[] text, int textLength, MobiBrowser browser) { 
        setFullScreenMode(true);
        this.url = url;
        this.label = label;
        this.text = text;
        this.textLength = textLength;
        this.browser = browser;
	screenWidth = getWidth()-1;
	screenHeight = getHeight();
        font = Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL);
        linkFont = Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_UNDERLINED, Font.SIZE_SMALL);
        fontHeight = font.getHeight() - OVERLAP;
        screenLines = screenHeight/fontHeight;
        spaceWidth = font.charWidth(' ');
        hyphenWidth = font.charWidth(Locale.HYPHEN);
        linePos = new int[screenLines+1];
        prevPageChar = '\n';
        index = new Hashtable();
        tags = new Tag();
        convertText();
        currTag = tags;
        browser.cache(url, this);
        if (currPagePos == 0) { 
            setPosition(pos);
        }
        repeatTimer = new Timer();
        addCommand(MobiBrowser.URL_CMD);
        addCommand(MobiBrowser.FORWARD_CMD);
        addCommand(MobiBrowser.BACKWARD_CMD);
        addCommand(MobiBrowser.BOOKMARKS_CMD);
        addCommand(MobiBrowser.HISTORY_CMD);
        addCommand(MobiBrowser.SEARCH_CMD);
        addCommand(MobiBrowser.REFRESH_CMD);
        addCommand(MobiBrowser.HELP_CMD);
        addCommand(MobiBrowser.QUIT_CMD);
        setCommandListener(this);
        Display.getDisplay(browser).setCurrent(this);
    }
    
    public void commandAction(Command c, Displayable d) 
    {
        Menu menu = browser.menu;
        menu.canvas = this;
        if (c == MobiBrowser.URL_CMD) { 
            new UrlForm(menu);
        } else if (c == MobiBrowser.FORWARD_CMD) { 
            browser.forward();
        } else if (c == MobiBrowser.BACKWARD_CMD) { 
            browser.backward();
        } else if (c == MobiBrowser.BOOKMARKS_CMD) { 
            new BookmarkList(menu);
        } else if (c == MobiBrowser.HISTORY_CMD) { 
            new HistoryList(menu);
        } else if (c == MobiBrowser.SEARCH_CMD) { 
            new SearchForm(menu);
        } else if (c == MobiBrowser.WEB_SEARCH_CMD) { 
            new WebSearchForm(menu);
        } else if (c == MobiBrowser.HELP_CMD) { 
            new HelpForm(menu);
        } else if (c == MobiBrowser.QUIT_CMD) { 
            browser.quit();
        } else if (c == MobiBrowser.REFRESH_CMD) { 
            new ConnectForm(url, currPagePos, browser);
        }
    }

    class RepeatTask extends TimerTask { 
        RepeatTask(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            keyAction(keyCode);
        }

        int keyCode;
    }

    void convertText() { 
        byte[] t = text;
        int length = textLength;
        int src, dst;
        int skip = 0;
        int preserve = 0;

        for (src = 0, dst = 0; src < length; src++) { 
            byte b = t[src];
            if (b == '<') { 
                if (src+3 < length && t[src+1] == '!' && t[src+2] == '-' && t[src+3] == '-')
                {
                    src += 3;
                    do { 
                        src += 1;
                    } while (src+3 < length && !(t[src] == '-' &&  t[src+1] == '-' &&  t[src+2] == '>'));
                    src += 2;
                    continue;
                }
                int start = src+1;
                while (++src < length && t[src] != '>') { 
                    if (t[src] == '<') { 
                        start = src+1;
                    }
                }
                String element = new String(t, start, src-start);
                String tagName = element.toLowerCase();
                for (int j = 0; j < skipTags.length; j++) { 
                    String st = skipTags[j];
                    if (tagName.equals(st) 
                        || (tagName.startsWith(st) && tagName.charAt(st.length()) == ' ')) 
                    {
                        skip += 1;
                        break;
                    } 
                }
                if (skip > 0 && tagName.startsWith("/")) { 
                    String suf = tagName.substring(1);
                    for (int j = 0; j < skipTags.length; j++) { 
                        if (suf.equals(skipTags[j])) { 
                            skip -= 1;
                            break;
                        } 
                    }
                }
                if (skip > 0) { 
                    continue;
                }
                if (tagName.startsWith("pre")) { 
                    preserve += 1;
                } else if (tagName.startsWith("/pre")) { 
                    preserve -= 1;
                } else if (tagName.startsWith("a href")) { 
                    String url = getUrl(element);
                    if (url.length() > 0 && !url.equals("#")) { 
                        tags.link(new Tag(url, dst));
                    }
                } else if (tagName.startsWith("a name")) { 
                    int qs = tagName.indexOf('"');
                    int qe = tagName.lastIndexOf('"');
                    if (tagName.substring(qs+1, qe).equals(label)) {
                        currPagePos = dst;
                    }
                    index.put(getUrl(element), new Integer(dst));
                } else if (tagName.equals("/a")) { 
                    if (tags.prev.end < 0) {
                        if (tags.prev.beg == dst) { 
                            tags.prev.unlink();
                        } else { 
                            tags.prev.end = dst;
                        }
                    }
                } else { 
                    for (int j = 0; j < breakTags.length; j++) { 
                        String bt = breakTags[j];
                        if (tagName.equals(bt) 
                            || (tagName.startsWith(bt) && tagName.charAt(bt.length()) == ' ')) 
                        { 
                            if (bt.equals("p")) { 
                                for (int i = 0; i < INDENTATION; i++) { 
                                    t[dst++] = ' ';
                                }
                            } else if (dst == 0 || t[dst-1] == '\n') { 
                                continue;
                            }
                            t[dst++] = '\n';                                
                            break;
                        }
                    }
                }
            } else if (skip == 0) { 
                if (b == '&') { 
                    int j, n = src + 8 < length ? src + 8 : length;
                    for (j = src+1; j < n && t[j] != ';'; j++);
                    if (j < n && t[j] == ';') { 
                        String symStr = new String(t, src+1, j-src-1).toLowerCase();
                        src = j;
                        char ch;
                        if (symStr.equals("amp")) { 
                            ch = '&';
                        } else if (symStr.equals("quot")) { 
                            ch = '"';
                        } else if (symStr.equals("lt")) { 
                            ch = '<';
                        } else if (symStr.equals("gt")) { 
                            ch = '>';
                        } else if (symStr.equals("nbsp")) { 
                            ch = ' ';
                        } else if (symStr.startsWith("#")) { 
                            ch = (char)Integer.parseInt(symStr.substring(1), 16);
                        } else { 
                            continue;
                        }
                        t[dst++] = (byte)ch;                        
                    } else { 
                        t[dst++] = '&';
                    }                    
                } else if (b == '\n' || b == ' ' || b == '\t') { 
                    if (preserve != 0) { 
                        t[dst++] = (byte)((b == '\n') ? '\r' : ' ');
                    } else { 
                        if (dst != 0  && t[dst-1] != ' ' && t[dst-1] != '\n') { 
                            t[dst++] = ' ';
                        }
                    }
                } else if (b != '\r') { 
                    t[dst++] = b;
                }
            }
        }
        textLength = dst;
    }
        
            

    public boolean search(String pattern) { 
        char[] p = pattern.toLowerCase().toCharArray();
        int n  = p.length;
        char[] q = new char[n];
        char prev = '\0';
        int  i = 0;
        int  pos = currPagePos;
      search:
        while (++pos < textLength) { 
            char ch = Locale.getChar(text[pos]);
            if (ch ==  ' ' && prev == ' ') { 
                continue;
            }
            prev = ch;
            if (i == n) { 
                i = 0;
            }
            q[i++] = Character.toLowerCase(ch);
            int j = 0;
            for (int k = i; k < n; k++, j++) { 
                if (p[j] != q[k]) { 
                    continue search;
                }
            }
            for (int k = 0; j < n; k++, j++) { 
                if (p[j] != q[k]) { 
                    continue search;
                }
            }
            setPosition(pos - n + 1);
            return true;
        }
        return false;
    }

    static void insertSpaces(StringBuffer buf, int nSpaces) { 
        if (nSpaces > 0) { 
            int i = 0;
            int n = buf.length();
            while (buf.charAt(i) == ' ') { 
                i += 1;
            }
            int nSeparators = 0;
            while (++i < n) { 
                if (buf.charAt(i) ==  ' ') { 
                    nSeparators += 1;
                }
            }
            if (nSeparators != 0) { 
                int ns1 = nSpaces / nSeparators;
                int ns2 = nSpaces % nSeparators;
                do { 
                    if (buf.charAt(--i) ==  ' ') { 
                        for (int j = 0; j < ns1; j++) { 
                            buf.insert(i, ' ');
                        }
                        if (ns2 != 0) { 
                            buf.insert(i, ' ');
                            ns2 -= 1;
                        }
                        nSeparators -= 1;
                    }
                } while (nSeparators != 0);
            }
        }
    }

    private static boolean isSeparator(char ch) { 
        return ch == ' ' || ch == '!' || ch == '.' || ch == ',' || ch == ';' || ch == '?' 
            || ch == '-' || ch == '+' || ch == '=' || ch == ')' || ch == '}' || ch == ']' 
            || ch == '\\' || ch == '/' || ch == ':' || ch == '>';
    }

    final char getChar(int i) { 
        if (i >= textLength) { 
            return '\0';
        }
        return Locale.getChar(text[i]);
    }

    int splitWord(StringBuffer buf, int n, int width, boolean align) { 
        char c1 = buf.charAt(n-1);
        char c2 = buf.charAt(n-2);
        if (isSeparator(c2)) { 
            if (c2 == ' ') { 
                buf.delete(n-2, buf.length());
                width -= spaceWidth + font.charWidth(c1);
            } else { 
                buf.delete(n-1, buf.length());
                width -= font.charWidth(c1);
            }
            if (align) { 
                insertSpaces(buf, (screenWidth - width)/spaceWidth);
            }
            return 1;
        }
        int i = Locale.split(buf, n);
        if (i <= 0) { 
            i = n-2;
        } 
        int rest = n-i;
        boolean insertHyphen = true;
        if (i >= 1 && isSeparator(buf.charAt(i-1))) { 
            if (buf.charAt(i-1) == ' ') { 
                i -= 1;
            }
            insertHyphen = false;
        } else if (i >= 2 && isSeparator(buf.charAt(i-2))) { 
            i -= buf.charAt(i-2) == ' ' ? 2 : 1; 
            rest += 1;
            insertHyphen = false;
        }
        for (int j = i; j < n; j++) { 
            width -= font.charWidth(buf.charAt(j));
        }
        buf.delete(i, buf.length());
        if (insertHyphen) { 
            buf.insert(i, Locale.HYPHEN);
            width += hyphenWidth;
        }
        if (align) { 
            insertSpaces(buf, (screenWidth - width)/spaceWidth);
        }
        return rest;
    }

    private static String getUrl(String tag) { 
        String url = tag.substring(tag.indexOf('=')+1).trim();
        if (url.startsWith("\"")) { 
            url = url.substring(1, url.indexOf('"', 1));
        } else { 
            int sep = url.indexOf(' ');
            if (sep >= 0) { 
                url = url.substring(0, sep);
            }
        }   
        int q = url.indexOf('?');
        if (q >= 0) { 
            String redirect = url.substring(q+1);
            if (redirect.startsWith("http://")) { 
                return redirect;
            }
        }
        return url;
    }

    protected void paint(Graphics g) {
        g.setColor(BACKGROUND_COLOR); 
        g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
        boolean containsCurrTag = false;

        while (true) { 
            int pos = currPagePos;
            boolean eof = false;
            Tag tag = tags.next;
            Tag firstPageTag = null;
            
            nLines = 0;     

            for (int i = 0; !eof && i < screenLines; i++) { 
                linePos[i] = pos;
                nLines += 1;
                int width = 0;
                StringBuffer buf = new StringBuffer();
                Tag firstTag = null;
                Tag lastTag = null;
                char ch = '\0';
                do { 
                    ch = getChar(pos);
                    while (tag != tags && tag.end <= pos) { 
                        tag = tag.next;
                    }
                    boolean isLink = false;
                    if (pos >= tag.beg && pos < tag.end) { 
                        isLink = false;
                        if (firstTag == null) { 
                            if (currTag.end <= currPagePos) { 
                                currTag = tag;
                            }
                            firstTag = lastTag = tag;
                            tag.pos = buf.length() - pos + tag.beg;
                        } else if (tag.beg == pos) {
                            lastTag = tag; 
                            tag.pos = buf.length();
                        }
                    }
                    pos += 1;
                    if (ch == ' ') { 
                        int j = pos-1;
                        while (--j >= 0 && text[j] == ' ');
                        if ((j < 0 || text[j] == '\r') && pos-j-1 <= INDENTATION) { 
                            buf.append(ch); 
                            width += spaceWidth;
                            continue;
                        } else if (j+2 < pos) { 
                            continue;
                        }
                    } else if (ch == '\r') {
                        ch = '\n';
                        if (pos < textLength && (text[pos] == ' ' || text[pos] == '\r')) { 
                            break;
                        }
                        ch =  ' ';
                    } else if (ch == '\0') {
                        eof = true;
                        break;
                    } else if (ch == '\n') {
                        break;
                    }
                    buf.append(ch);
                    width += font.charWidth(ch);
                } while (width <= screenWidth);
                
                //System.out.println("String='" + buf.toString() + "', width=" + width);
                
                if (width > screenWidth) { 
                    if (ch == ' ' || ch == '\n') { 
                        buf.deleteCharAt(buf.length()-1);
                        width -= spaceWidth;
                        if (firstTag == null) { 
                            insertSpaces(buf, (screenWidth - width) / spaceWidth);
                        }
                    } else { 
                        int n = buf.length();
                        if (!eof) { 
                            int p = pos;
                            while ((ch = getChar(p++)) > ' ') { 
                                buf.append(ch);
                            }
                        }
                        pos -= splitWord(buf, n, width, firstTag == null);
                    }
                }
                String str = buf.toString(); 
                int strLen = str.length();
                g.setColor(FOREGROUND_COLOR); 
                g.setFont(font);
                g.drawString(str, 0, i*fontHeight, Graphics.LEFT|Graphics.TOP);  
                if (firstTag != null) { 
                    if (firstPageTag == null) { 
                        firstPageTag = firstTag;
                    }
                    g.setFont(linkFont);
                    lastTag = lastTag.next;
                    do { 
                        int tagLen = firstTag.end - firstTag.beg;
                        int tagPos = firstTag.pos;
                        if (tagPos < 0) { 
                            tagLen += tagPos;
                            tagPos = 0;
                        }
                        if (tagPos < strLen) { 
                            if (tagPos + tagLen > strLen) {
                                tagLen = strLen - tagPos;
                            }
                            if (firstTag == currTag) { 
                                g.setColor(CURRENT_LINK_BACKGROUND_COLOR);
                                g.fillRect(font.substringWidth(str, 0, tagPos), i*fontHeight, 
                                           font.substringWidth(str, tagPos, tagLen), fontHeight);
                                g.setColor(CURRENT_LINK_FOREGROUND_COLOR);
                                containsCurrTag = true;
                            } else { 
                                g.setColor(firstTag.url.startsWith("#") || browser.historyHash.get(firstTag.url) != null
                                           ? VISITED_LINK_COLOR : LINK_COLOR);
                            }
                            // System.out.println("tagPos=" + tagPos + ", tagLen=" + tagLen + ", firstTag.beg=" + firstTag.beg + ", firstTag.end=" + firstTag.end + ", firstTag.pos=" + firstTag.pos + ", strLen=" + strLen);
                            g.drawSubstring(str, tagPos, tagLen, font.substringWidth(str, 0, tagPos), 
                                            i*fontHeight, Graphics.LEFT|Graphics.TOP);  
                        }
                    } while ((firstTag = firstTag.next) != lastTag);
                }
            }
            currPageEnd = pos;
            linePos[nLines] = pos;

            if (!containsCurrTag && firstPageTag != null) { 
                containsCurrTag = true;
                currTag = firstPageTag;
                continue;
            }
            break;
        }

        if (textLength != 0) { 
            int sliderLength = (currPageEnd - currPagePos)*screenHeight/textLength;
            if (sliderLength < MIN_SLIDER_LENGTH) { 
                sliderLength = MIN_SLIDER_LENGTH;
            }
            int sliderPos = currPagePos*screenHeight/textLength;
            g.setColor(SLIDER_COLOR);
            g.drawLine(screenWidth, 0, screenWidth, screenHeight);
            g.setColor(SLIDER_POS_COLOR);              
            g.drawLine(screenWidth, sliderPos, screenWidth, sliderPos+sliderLength);
        }        
    }

    void moveBackward(int shift) { 
         int pos = currPagePos;
         int width = 0;
         prevPageChar = '\n';
         while (--pos >= 0 && shift > 0) { 
             char ch = getChar(pos);
             if (ch == '\n') { 
                 width = 0;
                 shift -= 1;
                 prevPageChar = '\n';
             } else { 
                 width += font.charWidth(ch);
                 if (width > screenWidth) { 
                     width = 0;
                     shift -= 1;
                     prevPageChar = ' ';
                 }
             }
         }
         setPosition(pos+1);
    }    

    void moveForward(int shift) { 
        if (nLines == screenLines) { 
            if (shift > screenLines) {
                shift = screenLines;
            }
            setPosition(linePos[shift]);
        }
    }

    public void setPosition(int pos) { 
        currPagePos = pos;
        browser.setCurrentPosition(pos);
    }
        

    public void keyPressed(int keyCode) {
        if (repeatTask != null) { 
            repeatTask.cancel();
        }
        repeatTask = new RepeatTask(keyCode);
        repeatTimer.schedule(repeatTask, REPEAT_DELAY, REPEAT_PERIOD);
    }

    public void keyReleased(int keyCode) {
        keyAction(keyCode);
        if (repeatTask != null) { 
            repeatTask.cancel();
	    repeatTask = null;
        }
    }

    private void keyAction(int keyCode) {
        switch (keyCode) {
        case KEY_NUM1:
        case KEY_NUM2:
        case KEY_NUM3:
        case KEY_NUM4:
        case KEY_NUM5:
        case KEY_NUM6:
        case KEY_NUM7:
        case KEY_NUM8:
        case KEY_NUM9:
            moveForward(keyCode - KEY_NUM0);
            break;            
        case KEY_NUM0:
            new Menu(browser, this);
            return;
        case KEY_STAR:
            browser.backward();
            return;
        case KEY_POUND:
            browser.forward();
            return;
        default:
            switch (getGameAction(keyCode)) { 
            case LEFT:
                if (currTag.prev != tags) { 
                    currTag = currTag.prev;
                    if (currTag.beg < currPagePos) { 
                        setPosition(currTag.beg);
                    }
                } else { 
                    moveBackward(1);
                }
                break;
            case RIGHT:
                if (currTag.next != tags) { 
                    currTag = currTag.next;
                    if (currTag.beg >= currPageEnd) { 
                        setPosition(currTag.beg);
                    }
                } else { 
                    moveForward(1);
                }
                break;
            case UP:
                moveBackward(screenLines);
                break;
            case DOWN:
                moveForward(screenLines);
                break;
            case FIRE:
                if (currTag != tags) { 
                    String currUrl = currTag.url;
                    if (currUrl.startsWith("#")) {
                        Integer pos = (Integer)index.get(currUrl.substring(1));
                        if (pos != null) { 
                            setPosition(pos.intValue());  
                            break;
                        }
                    } else { 
                        if (currUrl.indexOf("//:") < 0) { 
                            if (currUrl.startsWith("/")) { 
                                currUrl = MobiBrowser.getHost(url) + currUrl;
                            } else { 
                                currUrl = MobiBrowser.getBaseUrl(url) + currUrl;
                            }
                        }
                        browser.load(currUrl, 0);
                    }
                }
                // no break;
            default:
                return;
            }
        }
        repaint();
    }
}
